using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Schema;
using System.Xml.XPath;

namespace Macke.KonfigurationNS
{
    /// <summary>
    /// Klasse zur Konfiguration von Anwendungen mittels XML.
    /// 
    /// Die Einstellungen werden in einer XML-Datei (Standard: default.xml)
    /// abgelegt und aus dieser geladen. Vernderte Eisntellungen werden
    /// in einer separaten Datei (Standard: config.xml) abgelegt.
    /// </summary>
    public class Konfiguration
    {
        
        #region Konstanten
     
        /// <summary>
        /// URI des zu verwendenden Schemas zur Validierung der XML-Dateien.
        /// </summary>
        private const string _XmlNamespace = "http://xml.stefanmacke.com/Schema/Konfiguration";
        
        /// <summary>
        /// Speicherort der .XSD-Datei des verwendeten Schemas.
        /// </summary>
        private const string _XmlSchemaLocation = "http://xml.stefanmacke.com/Schema/Konfiguration.xsd";

        /// <summary>
        /// Prfix, dass zur Identifikation des verwendeten Schemas genutzt wird.
        /// </summary>
        private const string _XmlNamespacePrefix = "k";

        /// <summary>
        /// Name des Wurzelelements der Konfigurationsdatei.
        /// </summary>
        private const string _XmlElementWurzel = "Konfiguration";

        /// <summary>
        /// Name der Bereichselemente der Konfigurationsdatei.
        /// </summary>
        private const string _XmlElementBereich = "Bereich";

        /// <summary>
        /// NAme der Einstellungselemente der Konfigurationsdatei.
        /// </summary>
        private const string _XmlElementEinstellung = "Einstellung";

        /// <summary>
        /// Name der Bezeichnungsattribute der Konfigurationsdatei.
        /// </summary>
        private const string _XmlElementBezeichnung = "Bezeichnung";

        /// <summary>
        /// Name der Beschreibungsattribute der Konfigurationsdatei.
        /// </summary>
        private const string _XmlElementBeschreibung = "Beschreibung";

        #endregion

        #region Attribute/Eigenschaften
        /// <summary>
        /// Der komplette Pfad der Datei fr die Standardeinstellungen
        /// des aktuellen Objekts.
        /// </summary>
        private string dateiDefault = null;

        /// <summary>
        /// Der komplette Pfad der Datei fr die Standardeinstellungen
        /// des aktuellen Objekts.
        /// </summary>
        private string DateiDefault
        {
            get
            {
                return this.dateiDefault;
            }
            set
            {
                if (this.ValidiereXmlDatei(value))
                {
                    this.dateiDefault = value;
                }
            }
        }

        /// <summary>
        /// Der komplette Pfad der Datei fr eigene Einstellungen des
        /// akuellen Obejkts.
        /// </summary>
        private string dateiConfig = null;

        /// <summary>
        /// Der komplette Pfad der Datei fr eigene Einstellungen des
        /// akuellen Obejkts.
        /// </summary>
        private string DateiConfig
        {
            get
            {
                return this.dateiConfig;
            }
            set
            {
                if (File.Exists(value))
                {
                    if (this.ValidiereXmlDatei(value))
                    {
                        this.dateiConfig = value;
                        this.xmlConfig = new XmlDocument();
                    }
                }
                else
                {
                    this.dateiConfig = value;
                }
            }
        }

        /// <summary>
        /// Die eingelesenen Standardeinstellungen.
        /// </summary>
        private XmlDocument xmlDefault = new XmlDocument();

        /// <summary>
        /// Die aktuellen eigenen Einstellungen.
        /// </summary>
        private XmlDocument xmlConfig = null;

        #endregion

        #region Konstruktoren
        /// <summary>
        /// Haupt-Konstruktor.
        /// </summary>
        /// <param name="xmlConfig">Kompletter Pfad der Datei fr die eigenen Konfigurationseinstellungen.</param>
        /// <param name="xmlDefault">Kompletter Pfad der Datei fr die Standard-Konfigurationseinstellungen.</param>
        public Konfiguration(string dateiConfig, string dateiDefault)
        {
            // Dateinamen setzen
            this.DateiDefault = dateiDefault;
            this.DateiConfig = dateiConfig;
            
            // Dateien einlesen
            this.DateienEinlesen();
        }

        /// <summary>
        /// Konstruktor mit Standardeinstellung fr die Standard-Konfigurationsdatei.
        /// </summary>
        /// <param name="xmlConfig">Kompletter Pfad der Datei fr die eigenen Konfigurationseinstellungen.</param>
        public Konfiguration(string dateiConfig)
            : this(dateiConfig, Path.Combine(Application.StartupPath, "default.xml"))
        { }

        /// <summary>
        /// Konstruktor mit Standardeinstellungen fr beide Konfigurationsdateien.
        /// </summary>
        public Konfiguration()
            : this(Path.Combine(Application.StartupPath, "config.xml"), Path.Combine(Application.StartupPath, "default.xml"))
        { }
        #endregion

        #region Structs
        /// <summary>
        /// Struktur einer Einstellung: Name, Pfad-Notation, Standard- und eigene Einstellung.
        /// </summary>
        public struct Einstellung
        {
            /// <summary>
            /// Bezeichnung der Einstellung. Diese wird im Konfigurationsmen angezeigt.
            /// </summary>
            public string Bezeichnung;
            
            /// <summary>
            /// Pfadnotation der Einstellung ("Bereich.Unterbereich.Einstellung").
            /// </summary>
            public string Pfad;

            /// <summary>
            /// Der Wert der Einstellung, der aus der Standard-Konfigurationsdatei eingelesen wurde.
            /// </summary>
            public string StandardWert;

            /// <summary>
            /// Der Wert der Einstellung, der aus der eigenen Konfigurationsdatei eingelesen bzw.
            /// neu gesetzt wurde.
            /// </summary>
            public string EigenerWert;

            /// <summary>
            /// Beschreibung der Funktion des Einstellung.
            /// </summary>
            public string Beschreibung;
        }

        /// <summary>
        /// Struktur eines Bereichs: Name, Pfad-Notation und Einstellungen.
        /// </summary>
        public struct Bereich
        {
            /// <summary>
            /// Bezeichnung des Bereichs. Diese wird im Konfigurationsmen angezeigt.
            /// </summary>
            public string Bezeichnung;
            
            /// <summary>
            /// Pfadnotation des Bereichs ("Bereich.Unterbereich").
            /// </summary>
            public string Pfad;
            
            /// <summary>
            /// Liste der zum Bereich gehrenden Einstellungen.
            /// </summary>
            public List<Einstellung> Einstellungen;

            /// <summary>
            /// Beschreibung des Bereichs.
            /// </summary>
            public string Beschreibung;
        }
        #endregion

        #region Methoden
        /// <summary>
        /// Prft die angegebene Datei auf Existenz und gltiges XML und 
        /// validiert sie gegen das Schema _XmlSchemaLocation.
        /// </summary>
        /// <param name="datei">Name der zu validierenden Datei.</param>
        /// <returns>True wenn erfolgreich validiert.</returns>
        private bool ValidiereXmlDatei(string datei)
        {
            XmlReader xmlReader = null;
            try
            {
                XmlReaderSettings xmlSettings = new XmlReaderSettings();
                xmlSettings.ValidationType = ValidationType.Schema;
                xmlSettings.Schemas.Add(_XmlNamespace, _XmlSchemaLocation);
                xmlReader = XmlReader.Create(datei, xmlSettings);
                while (xmlReader.Read())
                { }
                return true;
            }
            catch (IOException ex1)
            {
                string fehler = "Datei konnte nicht geladen werden:\n" + ex1;
                throw new IOException(fehler, ex1);
            }
            catch (XmlException ex2)
            {
                string fehler = "XML-Fehler beim Laden der Datei " + datei + ":\n" + ex2;
                throw new XmlException(fehler, ex2);
            }
            catch (XmlSchemaException ex3)
            {
                string fehler = "Fehler beim Validieren der Datei " + datei + ":\n" + ex3;
                throw new XmlSchemaException(fehler, ex3);
            }
            catch (Exception ex)
            {
                throw ex;
            }
            finally
            {
                if (xmlReader != null)
                {
                    xmlReader.Close();
                }
            }
        }

        /// <summary>
        /// Gibt den Wert der Einstellung des angegebenen Pfades zurck.
        /// </summary>
        /// <param name="einstellung">Die gewnschte Einstellung in "Klassennotation" (z.B. "Endpunkte.Dateiformat").</param>
        /// <returns>Wert der gewnschten Einstellung.</returns>
        public string GetEinstellung(string einstellung)
        {
            // Vorbereiten des XPath-Ausdrucks und anderer genutzer Variablen
            string xpathEinstellung = this.XPathZurEinstellung(einstellung);
            string wertDerEinstellung = "";
            bool einstellungGefunden = false;
            Exception ex = null;

            // zunchst die eigenen Einstellungen ausgeben, falls gesetzt
            if (this.xmlConfig != null)
            {
                try
                {
                    wertDerEinstellung = this.LeseEinstellung(this.xmlConfig, xpathEinstellung).Value;
                    einstellungGefunden = true;
                }
                catch (Exception e)
                {
                    ex = e;
                    einstellungGefunden = false;
                }
            }

            // falls Einstellung nicht selbst definiert, Standard zurckgeben
            if (!einstellungGefunden)
            {
                try
                {
                    wertDerEinstellung = this.LeseEinstellung(this.xmlDefault, xpathEinstellung).Value;
                    einstellungGefunden = true;
                }
                catch (Exception e)
                {
                    ex = e;
                    einstellungGefunden = false;
                }
            }

            // Ausnahme werfen, wenn der Wert in keiner der Dateien gefunden werden kann
            if (!einstellungGefunden)
            {
                string fehler = "Einstellung konnte weder in den Standard- noch in den eigenen Einstellungen gefunden werden:\n" + einstellung + "\n" + xpathEinstellung.ToString();
                throw new IndexOutOfRangeException(fehler, ex);
            }

            return wertDerEinstellung;
        }

        /// <summary>
        /// Setzt die angegebene Einstellung auf den bergebenen Wert.
        /// </summary>
        /// <param name="einstellung">Die gewnschte Einstellung in "Klassennotation" (z.B. "Endpunkte.Dateiformat").</param>
        /// <param name="wert">Der zu setzende Wert fr die bergebene Einstellung.</param>
        public void SetEinstellung(string einstellung, string wert)
        {
            string xpathEinstellung = this.XPathZurEinstellung(einstellung);

            // zuerst herausfinden, ob die Einstellung berhaupt gesetzt werden darf
            try
            {
                this.LeseEinstellung(this.xmlDefault, xpathEinstellung);
            }
            catch (IndexOutOfRangeException exi)
            {
                throw new IndexOutOfRangeException("Einstellung nicht in Standard-Konfigurationsdatei gesetzt:\nEinstellung: " + einstellung + "\nDatei: " + this.dateiDefault, exi);
            }

            // Einstellung in eigener Konfiguration setzen
            // schon gesetzt?
            try
            {
                this.LeseEinstellung(this.xmlConfig, xpathEinstellung).SetValue(wert);
            }
            catch
            {
                this.ErstelleEinstellung(einstellung, wert);
            }

        }

        /// <summary>
        /// Konvertiert die bergebene Einstellung in "Klassennotation"
        /// in einen gltigen XPath-Ausdruck.
        /// </summary>
        /// <param name="einstellung">Die Einstellung in "Klassennotation" (z.B. "Endpunkte.Dateiformat").
        /// Der Wert kann auch mit einem "." enden, dann wird nur der entsprechende Bereich umgeformt.</param>
        /// <returns>XPath-Ausdruck zur angegebenen Einstellung.</returns>
        private string XPathZurEinstellung(string einstellung)
        {
            // Vorbereiten des XPath-Ausdrucks und anderer genutzer Variablen
            this.PruefeFormatDerEinstellung(einstellung);
            StringBuilder xpathEinstellung = new StringBuilder("/" + _XmlNamespacePrefix + ":" + _XmlElementWurzel + "/");

            // String-Verarbeitung/-Umformung in XPath-Ausdruck
            string[] teile = new Regex(@"\.").Split(einstellung);
            for (int i = 0; i < teile.Length; i++)
            {
                // alle Teile bis auf den letzten sind Bereiche
                if (i < (teile.Length - 1))
                {
                    xpathEinstellung.Append(_XmlNamespacePrefix + ":" + _XmlElementBereich + "[@" + _XmlNamespacePrefix + ":" + _XmlElementBezeichnung + "='" + teile[i] + "']/");
                }
                else
                {
                    // wenn ein kompletter Pfad bergeben wird, ist das letzte Element eine Einstellung...
                    if (!einstellung.EndsWith("."))
                    {
                        xpathEinstellung.Append(_XmlNamespacePrefix + ":" + _XmlElementEinstellung + "[@" + _XmlNamespacePrefix + ":" + _XmlElementBezeichnung + "='" + teile[i] + "']");
                    }
                    // ...sonst ein Bereich (Notation endet mit ".")
                    else
                    {
                        // abschlieenden Slash entfernen
                        xpathEinstellung.Remove(xpathEinstellung.Length - 1, 1);
                    }
                }
            }

            return xpathEinstellung.ToString();
        }

        /// <summary>
        /// Sucht den angegebenen Knoten im bergeben XMLDocument und gibt
        /// den Wert zurck.
        /// </summary>
        /// <param name="xmldoc">XMLDocument, in dem gesucht werden soll.</param>
        /// <param name="xpath">XPath-Ausdruck des zu suchenden Knotens.</param>
        /// <returns>Der gesuchte Knoten als XPathNavigator.</returns>
        private XPathNavigator LeseEinstellung(XmlDocument xmldoc, string xpath)
        {
            // Navigator vorbereiten
            XPathNavigator nav = xmldoc.CreateNavigator();
            XPathExpression exp = nav.Compile(xpath);

            // Namespaces vorbereiten
            XmlNamespaceManager mgr = new XmlNamespaceManager(xmldoc.NameTable);
            mgr.AddNamespace(_XmlNamespacePrefix, _XmlNamespace);
            exp.SetContext(mgr);

            // Knoten selektieren
            XPathNodeIterator iter = nav.Select(exp);
            
            // Knoten gefunden?
            if (iter.Count == 1)
            {
                // von root zum ausgewhlten Knoten wechseln
                iter.MoveNext();
                return iter.Current;
            }
            else
            {
                if (iter.Count == 0)
                {
                    throw new IndexOutOfRangeException("Knoten nicht gefunden: " + xpath);
                }
                else
                {
                    throw new IndexOutOfRangeException("Mehrdeutiger Ausdruck: " + xpath);
                }
            }
        }

        /// <summary>
        /// Erstellt ein Objekt vom Typ XmlDocument, dass die eigenen Einstellungen
        /// enthlt.
        /// </summary>
        private void ErstelleDateiConfig()
        {
            // nur erstellen, wenn noch nicht vorhanden
            if (this.xmlConfig == null)
            {
                // XmlDocument etc. instantiieren
                XmlDocument xmlDoc = new XmlDocument();
                XmlDeclaration xmlDeclaration = xmlDoc.CreateXmlDeclaration("1.0", "UTF-8", null);
                XmlElement rootNode = xmlDoc.CreateElement(_XmlNamespacePrefix + ":" + _XmlElementWurzel,_XmlNamespace);
                
                // eigener Namespace
                XmlAttribute attr = xmlDoc.CreateAttribute("xmlns:" + _XmlNamespacePrefix);
                attr.InnerText = _XmlNamespace;
                rootNode.Attributes.Append(attr);
                
                // Schema-Namespace
                attr = xmlDoc.CreateAttribute("xmlns:xsi");
                attr.InnerText = "http://www.w3.org/2001/XMLSchema-instance";
                rootNode.Attributes.Append(attr);

                // Schema-Location
                attr = xmlDoc.CreateAttribute("xsi:schemaLocation", "http://www.w3.org/2001/XMLSchema-instance");
                attr.InnerText = _XmlNamespace + " " + _XmlSchemaLocation;
                rootNode.Attributes.Append(attr);

                // Deklaration und Wurzelelement dem XmlDocument hinzufgen
                xmlDoc.InsertBefore(xmlDeclaration, xmlDoc.DocumentElement);
                xmlDoc.AppendChild(rootNode);
                this.xmlConfig = xmlDoc;
            }
        }

        /// <summary>
        /// Schreibt das angegebene XmlDocument in eine XML-Datei.
        /// </summary>
        private void SchreibeConfigDatei()
        {
            // Einstellungen mit den Standardwerten vergleichen
            // Navigator vorbereiten
            XPathNavigator nav = this.xmlConfig.CreateNavigator();
            
            // alle Einstellungselemente auswhlen
            XPathExpression exp = nav.Compile("//" + _XmlNamespacePrefix + ":" + _XmlElementEinstellung);

            // Namespaces vorbereiten
            XmlNamespaceManager mgr = new XmlNamespaceManager(this.xmlConfig.NameTable);
            mgr.AddNamespace(_XmlNamespacePrefix, _XmlNamespace);
            exp.SetContext(mgr);

            // Knoten selektieren
            XPathNodeIterator iter = nav.Select(exp);
            
            // Knoten gefunden?
            while (iter.MoveNext())
            {
                string notation = this.EinstellungZumElement(iter.Current);
                
                // Werte vergleichen
                string einstellungConfig = this.LeseEinstellung(this.xmlConfig, this.XPathZurEinstellung(notation)).Value;
                string einstellungDefault = this.LeseEinstellung(this.xmlDefault, this.XPathZurEinstellung(notation)).Value;


                // gleiche Einstellungen aus config.xml entfernen
                if (einstellungConfig == einstellungDefault)
                {
                    XPathNavigator deleteNode = this.LeseEinstellung(this.xmlConfig, this.XPathZurEinstellung(notation));
                    deleteNode.DeleteSelf();
                    // bergeordnete Bereiche lschen, wenn sie leer sind
                    while (!deleteNode.HasChildren && (deleteNode.LocalName != _XmlElementWurzel))
                    {
                        deleteNode.DeleteSelf();
                    }
                    // Knoten nach Lschen neu einlesen
                    iter = nav.Select(exp);
                }
            }

            // wenn Elemente vorhanden sind, Datei schreiben...
            if (this.xmlConfig.DocumentElement.HasChildNodes)
            {
                XmlTextWriter writer = new XmlTextWriter(this.DateiConfig, Encoding.UTF8);
                writer.Formatting = Formatting.Indented;
                writer.WriteStartDocument();
                foreach (XmlNode node in this.xmlConfig.ChildNodes)
                {
                    if (node.LocalName != "xml")
                    {
                        this.SchreibeXmlNode(writer, node);
                    }
                }
                writer.WriteEndDocument();
                writer.Close();
            }
            // ...sonst lschen
            else
            {
                if (File.Exists(this.dateiConfig))
                {
                    try
                    {
                        File.Delete(this.dateiConfig);
                    }
                    catch
                    { 
                    }
                }
            }
        }

        /// <summary>
        /// Gibt das bergebene Element auf dem bergebenen XmlTextWriter aus.
        /// Dabei werden alle Attribute und Kindelemente (mittels Rekursion) 
        /// mit ihren qualifizierten Namen ausgegeben.
        /// </summary>
        /// <param name="writer">XmlTextWriter, der das XmlDocument schreiben soll.</param>
        /// <param name="node">Element, das auf dem XmlTextWriter ausgegeben werden soll.</param>
        private void SchreibeXmlNode(XmlTextWriter writer, XmlNode node)
        {
            // Element ffnen
            writer.WriteStartElement(node.Prefix, node.LocalName, node.NamespaceURI);
            
            // Attribute schreiben
            if (node.Attributes != null)
            {
                foreach (XmlAttribute attr in node.Attributes)
                {
                    writer.WriteAttributeString(attr.Prefix, attr.LocalName, attr.NamespaceURI, attr.Value.ToString());
                }
            }
            
            // Kind-Elemente schreiben
            if (node.HasChildNodes)
            {
                foreach (XmlNode childNode in node.ChildNodes)
                {
                    if (childNode.NodeType != XmlNodeType.Text)
                    {
                        this.SchreibeXmlNode(writer, childNode);
                    }
                    // wenn Typ = Text, ist der Wert des Elements erreicht
                    // diesen einfach anhngen
                    else
                    {
                        writer.WriteString(childNode.InnerText);
                    }
                }
            }

            // Element schlieen
            writer.WriteEndElement();
        }

        /// <summary>
        /// Erstellt die bergebene Einstellung in der eigenen Konfiguration.
        /// </summary>
        /// <param name="einstellung">Die zu erstellende Einstellung in "Klassennotation" (z.B. "Endpunkte.Dateiformat").</param>
        /// <param name="wert">Wert, der der neuen Einstellung zugewiesen wird.</param>
        private void ErstelleEinstellung(string einstellung, string wert)
        {
            this.PruefeFormatDerEinstellung(einstellung);

            // Dokument erstellen, wenn es noch nicht existiert
            if (this.xmlConfig == null)
            {
                this.ErstelleDateiConfig();
            }

            // Einstellung in Bereiche und Einstellung aufteilen
            string[] teile = new Regex(@"\.").Split(einstellung);

            // Namespaces vorbereiten
            XmlNamespaceManager mgr = new XmlNamespaceManager(this.xmlConfig.NameTable);
            mgr.AddNamespace(_XmlNamespacePrefix, _XmlNamespace);
            
            // Element, dem ein neu zu erstellendes hinzugefgt wird
            XmlElement parent = this.xmlConfig.DocumentElement;

            // auch jeden bergeordneten Breich erstellen, falls nicht vorhanden
            for (int i = 0; i < teile.Length; i++)
            {
                // aktuelles Element bestimmen
                StringBuilder aktuellesElement = new StringBuilder();
                for (int j = 0; j <= i; j++)
                {
                    aktuellesElement.Append(teile[j]);
                    if (j < teile.Length - 1)
                    {
                        aktuellesElement.Append(".");
                    }
                }

                // prfen, ob aktuelles Element vorhanden ist
                XmlNode element = this.xmlConfig.SelectSingleNode(this.XPathZurEinstellung(aktuellesElement.ToString()), mgr);
                
                // Element erstellen, wenn nicht vorhanden
                if (element == null)
                {
                    // Bereichselemente erstellen
                    XmlElement neuesElement = null;
                    if (i < (teile.Length - 1))
                    {
                        neuesElement = this.xmlConfig.CreateElement(_XmlNamespacePrefix, _XmlElementBereich, _XmlNamespace);
                        neuesElement.SetAttribute(_XmlElementBezeichnung, _XmlNamespace, teile[i]);
                    }
                    // Einstellungselement erstellen
                    else
                    {
                        neuesElement = this.xmlConfig.CreateElement(_XmlNamespacePrefix, _XmlElementEinstellung, _XmlNamespace);
                        neuesElement.SetAttribute(_XmlElementBezeichnung, _XmlNamespace, teile[i]);
                        neuesElement.InnerText = wert;
                    }
                    // neues Element an Elternelement anhngen
                    parent.AppendChild(neuesElement);

                    // auf die nchsttiefere Ebene wechseln
                    parent = neuesElement;
                }
                // Element ist vorhanden
                else
                {
                    // auf die nchsttiefere Ebene wechseln
                    parent = (XmlElement)element;
                }
            }

        }

        /// <summary>
        /// Prft das Format der bergebenen Einstellung in Klassennotation.
        /// </summary>
        /// <param name="einstellung">Der zu prfende String in Klassennotation.</param>
        private void PruefeFormatDerEinstellung(string einstellung)
        {
            // Format des bergebenen Strings prfen
            Regex stringformat = new Regex(@"^[a-zA-Z0-9\.]*$");
            if (!stringformat.Match(einstellung).Success)
            {
                throw new FormatException("Der bergebene String hat das falsche Format: " + einstellung);
            }
        }

        /// <summary>
        /// Speichert die aktuelle Konfiguration in config.xml.
        /// </summary>
        public void Speichern()
        {
            if (this.xmlConfig != null)
            {
                this.SchreibeConfigDatei();
            }
        }

        /// <summary>
        /// Gibt die Klassennotation ("Bereich.Einstellung") zum
        /// bergebenen Element zurck.
        /// </summary>
        /// <param name="node">Element, dessen Klassennotation gesucht ist.</param>
        /// <returns>Klassennotation des bergebenen Elements ("Bereich.Einstellung").</returns>
        private string EinstellungZumElement(XPathNavigator node)
        {
            // Bezeichnung des Elements auslesen
            StringBuilder notation = new StringBuilder(node.GetAttribute(_XmlElementBezeichnung, _XmlNamespace));
            
            // alle Bereichselemente von hinten nach vorne hinzufgen
            while (node.MoveToParent())
            {
                if (node.LocalName == _XmlElementBereich)
                {
                    notation.Insert(0, node.GetAttribute(_XmlElementBezeichnung, _XmlNamespace) + ".");
                }
            }

            return notation.ToString();
        }

        /// <summary>
        /// Gibt die Liste mit Bereichen unterhalb des Wurzelelements zurck.
        /// </summary>
        /// <returns>Liste mit Bereichen unterhalb des Wurzelelements.</returns>
        public List<Bereich> GetBereiche()
        {
            return this.GetBereiche("");
        }
        
        /// <summary>
        /// Gibt die Liste mit Bereichen unterhalb des bergebenen Bereichs zurck.
        /// </summary>
        /// <param name="bereich">Pfad des Bereichs ("Bereich.Unterbereich") OHNE abschlieenden Punkt.</param>
        /// <returns>Liste der Unterbereiche zum bergebenen Bereich.</returns>
        public List<Bereich> GetBereiche(string pfad)
        {
            // XPath-Ausdruck erstellen
            string xpath = null;
            
            // Bereich bergeben
            if (pfad.Length > 0)
            {
                xpath = this.XPathZurEinstellung(pfad + ".") + "/" + _XmlNamespacePrefix + ":" + _XmlElementBereich;
            }
            // kein Bereich bergeben: Wurzelelement verwenden
            else
            {
                xpath = "/" + _XmlNamespacePrefix + ":" + _XmlElementWurzel + "/" + _XmlNamespacePrefix + ":" + _XmlElementBereich;
            }

            // Bereiche und Abfrage vorbereiten
            List<Bereich> bereiche = new List<Bereich>();
            XPathNavigator nav = this.xmlDefault.CreateNavigator();
            XPathExpression exp = nav.Compile(xpath);

            // Namespaces vorbereiten
            XmlNamespaceManager mgr = new XmlNamespaceManager(this.xmlDefault.NameTable);
            mgr.AddNamespace(_XmlNamespacePrefix, _XmlNamespace);
            exp.SetContext(mgr);

            // Knoten selektieren
            XPathNodeIterator iter = nav.Select(exp);
            
            // Knoten gefunden?
            while (iter.MoveNext())
            {
                // Bereich erstellen
                Bereich bereich = new Bereich();
                bereich.Bezeichnung = iter.Current.GetAttribute(_XmlElementBezeichnung, _XmlNamespace);
                bereich.Pfad = this.EinstellungZumElement(iter.Current);
                
                // Bereich zur Liste hinzufgen
                bereiche.Add(bereich);
            }

            return bereiche;
        }

        /// <summary>
        /// Gibt den Bereich zum angegeben Pfad zurck.
        /// </summary>
        /// <param name="pfad">Pfad des Bereichs ("Bereich.Unterbereich") OHNE abschlieenden Punkt.</param>
        /// <returns>Bereich zum angegeben Pfad.</returns>
        public Bereich GetBereich(string pfad)
        {
            if (pfad.Length == 0)
            {
                throw new InvalidDataException("Der bergebene Pfad ist leer!");
            }

            // XPath-Ausdruck erstellen
            string xpath = this.XPathZurEinstellung(pfad + ".");

            // Abfrage vorbereiten
            XPathNavigator nav = this.xmlDefault.CreateNavigator();
            XPathExpression exp = nav.Compile(xpath);

            // Namespaces vorbereiten
            XmlNamespaceManager mgr = new XmlNamespaceManager(this.xmlDefault.NameTable);
            mgr.AddNamespace(_XmlNamespacePrefix, _XmlNamespace);
            exp.SetContext(mgr);

            // Knoten selektieren
            XPathNodeIterator iter = nav.Select(exp);

            // Knoten gefunden?
            if (iter.Count != 1)
            {
                throw new IndexOutOfRangeException("Bereich nicht gefunden oder mehrdeutig:\n" + pfad + "\n" + xpath);
            }
            // vom Wurzelelement zum gefundenen wechseln
            iter.MoveNext();

            // Bereich erstellen
            Bereich bereich = new Bereich();
            bereich.Bezeichnung = iter.Current.GetAttribute(_XmlElementBezeichnung, _XmlNamespace);
            bereich.Beschreibung = iter.Current.GetAttribute(_XmlElementBeschreibung, _XmlNamespace);
            bereich.Pfad = this.EinstellungZumElement(iter.Current);
            bereich.Einstellungen = new List<Einstellung>();

            // Einstellungen ermitteln
            exp = nav.Compile(this.XPathZurEinstellung(bereich.Pfad + ".") + "/" + _XmlNamespacePrefix + ":" + _XmlElementEinstellung);
            exp.SetContext(mgr);
            XPathNodeIterator iterEinstellungen = nav.Select(exp);
            while (iterEinstellungen.MoveNext())
            {
                // Einstellung erstellen
                Einstellung einstellung = new Einstellung();
                einstellung.Bezeichnung = iterEinstellungen.Current.GetAttribute(_XmlElementBezeichnung, _XmlNamespace);
                einstellung.Beschreibung = iterEinstellungen.Current.GetAttribute(_XmlElementBeschreibung, _XmlNamespace);
                einstellung.Pfad = this.EinstellungZumElement(iterEinstellungen.Current);
                einstellung.StandardWert = this.LeseEinstellung(this.xmlDefault, this.XPathZurEinstellung(einstellung.Pfad)).Value;
                // eigene Einstellung lesen. wenn nicht vorhanden, Standard nutzen
                try
                {
                    einstellung.EigenerWert = this.LeseEinstellung(this.xmlConfig, this.XPathZurEinstellung(einstellung.Pfad)).Value;
                }
                catch
                {
                    einstellung.EigenerWert = einstellung.StandardWert;
                }
                bereich.Einstellungen.Add(einstellung);
            }
            
            return bereich;
        }

        /// <summary>
        /// Zeigt das Formular zum Bearbeiten der Konfiguration an.
        /// </summary>
        public void ZeigeKonfigurationsFormular()
        {
            new FormKonfiguration(this).ShowDialog();
        }

        /// <summary>
        /// Liest die beiden Konfigurationsdateien ein.
        /// </summary>
        public void DateienEinlesen()
        {
            this.xmlDefault.Load(this.DateiDefault);
            if (this.xmlConfig != null)
            {
                this.xmlConfig.Load(this.DateiConfig);
            }
        }

        #endregion
    }
}
